# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016, William Brown <william at blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import ldap

from idm389.organisationalunit import OrganisationalUnits

from lib389.plugins import Plugins

def _init_config(inst, basedn, log, commit=True):
    # * Set password hash to ssha512
    if commit:
        inst.config.set('passwordStorageScheme', 'SSHA512')
    elif inst.config.get_attr_val('passwordStorageScheme') != 'SSHA512':
        log.info('Would update passwordStorageScheme')
    if commit:
        inst.config.set('nsslapd-rootpwstoragescheme', 'SSHA512')
    elif inst.config.get_attr_val('nsslapd-rootpwstoragescheme') != 'SSHA512':
        log.info('Would update nsslapd-rootpwstoragescheme')
    # * Allow password migration.
    if commit:
        inst.config.set('nsslapd-allow-hashed-passwords', 'on')
    elif inst.config.get_attr_val('nsslapd-allow-hashed-passwords') != 'on':
        log.info('Would update nsslapd-allow-hashed-passwords')
    # * enable audit log
    if commit:
        inst.config.enable_log('audit')
    elif inst.config.get_attr_val('nsslapd-%slog-logging-enabled' % 'audit') != 'on':
        log.info('Would update nsslapd-%slog-logging-enabled' % 'audit')
    # * enable change log?
    # * 
    # * Turn on certain indexes
    #
    #
    #
    return True

def _init_schema(inst, basedn, log, commit=True):
    # We need to fix the ldap ssh public key schema here.
    attr = "( 1.3.6.1.4.1.24552.500.1.1.1.13 NAME 'sshPublicKey' DESC 'MANDATORY: OpenSSH Public key' EQUALITY octetStringMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.40 )"
    oc = "( 1.3.6.1.4.1.24552.500.1.1.2.0 NAME 'ldapPublicKey' SUP top AUXILIARY DESC 'MANDATORY: OpenSSH LPK objectclass' MUST ( uid ) MAY ( sshPublicKey ) )"

    result = inst.schema.query_attributetype('ntuserntpassword')
    if result is None:
        log.info('You must upgrade your Directory Server install. Missing ntuserntpassword')
        return False

    attrtype = inst.schema.query_attributetype('sshPublicKey')
    if attrtype is None:
        if commit:
            inst.schema.add_schema('attributeTypes', attr)
        else:
            log.info('Would add sshPublicKey attributetypes')

    result = inst.schema.query_objectclass('ldapPublicKey')
    if result is None:
        if commit:
            inst.schema.add_schema('objectClasses', oc)
        else:
            log.info('Would add ldapPublicKey objectclasses')

    return True

def _init_plugins(inst, basedn, log, commit=True):
    # We need:
    # * Managed Entries

    # * uiduniq
    ## Will need two instances, one for uid, one for mail.
    uiduniq = inst.plugins.get('attribute uniqueness')
    if commit:
        uiduniq.enable()
    elif not uiduniq.status():
        log.info('Would enable uid uniqueness plugin')
    # * refint
    refintplugin = inst.plugins.get('referential integrity postoperation')
    if commit:
        refintplugin.enable()
    elif not refintplugin.status():
        log.info('Would enable referential integrity plugin')
    # * cos
    cosplugin = inst.plugins.get('Class of Service')
    if commit:
        cosplugin.enable()
    elif not cosplugin.status():
        log.info('Would enable Class of Service plugin')
    # * roles ?
    rolesplugin = inst.plugins.get('Roles Plugin')
    if commit:
        rolesplugin.enable()
    elif not rolesplugin.status():
        log.info('Would enable Roles Plugin')
    return True

def _init_containers(inst, basedn, log, commit=True):
    # We need:
    ous = OrganisationalUnits(inst, basedn)

    required = [
        ('Groups', 'Structure containing groups'),
        ('People', 'Accounts which map to real identites'),
        ('Services', 'Computer Service accounts which request DS bind'),
        ('Permissions', 'Members of these groups, have a permission in the Directory'),
        # Do I need roles also?
    ]
    # Under what conditions can this fail?
    for name, desc in required:
        try:
            ou = ous.get(name)
        except ldap.NO_SUCH_OBJECT:
            if commit is True:
                log.info('Creating ou=%s,%s' % (name, basedn) )
                properties = {
                    'ou' : name,
                    'description': desc,
                }
                ous.create(properties=properties)
            else:
                log.info('Would create ou=%s,%s' % (name, basedn))
    return True


def _init_permissions(inst, basedn, log, commit=True):
    # This will add the needed permissions for correct operation
    # Should we purge existing acis?
    # I don't think so ....

    # Create the base acis
    # Create the base permissions.

    return True

def _init_roles(inst, basedn, log, commit=True):
    return True

def _init(inst, basedn, log, commit=False):
    if not _init_config(inst, basedn, log, commit=commit):
        log.error("Config migration check failed")
        return False

    if not _init_schema(inst, basedn, log, commit=commit):
        log.error("Schema migration check failed")
        return False

    if not _init_plugins(inst, basedn, log, commit=commit):
        log.error("Plugin migration check failed")
        return False

    if not _init_containers(inst, basedn, log, commit=commit):
        log.error("Container migration check failed")
        return False

    if not _init_permissions(inst, basedn, log, commit=commit):
        log.error("Permissions migration check failed")
        return False

    if not _init_roles(inst, basedn, log, commit=commit):
        log.error("Roles migration check failed")
        return False
    return True

