#!/usr/bin/python3

# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016, William Brown <william at blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

# We need this for error types
import ldap

from lib389._mapped_object import DSLdapObject, DSLdapObjects

from idm389.core import _init

from idm389.group import Group
from idm389.posixgroup import PosixGroup, PosixGroups
from idm389.user import UserAccount, UserAccounts

# These are definition of IPA object types we want to copy.

class IpaUser(DSLdapObject):
    def __init__(self, instance, dn=None, batch=False):
        super(IpaUser, self).__init__(instance, dn, batch)
        self._rdn_attribute = 'uid'
        self._must_attributes = []
        self._create_objectclasses = [
            'ipaobject',
            'person',
            'top',
            'ipasshuser',
            'inetorgperson',
            'organizationalperson',
            'krbticketpolicyaux',
            'krbprincipalaux',
            'inetuser',
            'posixaccount',
            'ipaSshGroupOfPubKeys',
            'mepOriginEntry',
            'ipantuserattrs',
        ]
        self._protected = True

    # 'ipasshuser', ipantuserattrs

    def convert_to_posixaccount(self):
        # Get the complete entry

        properties = {}
        attribute_map = [
            # (ipa -> ds)
            ('uid', 'uid'),
            ('displayName', 'displayName'),
            ('loginShell', 'loginShell'),
            ('initials', 'initials'),
            ('gecos', 'gecos'),
            ('sn', 'sn'),
            ('homeDirectory', 'homeDirectory'),
            ('mail', 'mail'),
            ('givenName', 'givenName'),
            ('cn', 'cn'),
            ('gidNumber', 'gidNumber'),
            ('uidNumber', 'uidNumber'),
            ('userPassword', 'userPassword'),
            ('title', 'title'),
            ('mobile', 'mobile'),
            ('telephoneNumber', 'telephoneNumber'),
            ('description', 'description'),
            ('initials', 'initials'),
            ('displayName', 'displayName'),
            ('postalCode', 'postalCode'),
            ('street', 'street'),
            ('st', 'st'),
            ('l', 'l'),
            ('ipaSshPubKey', 'sshPublicKey'),
        ]

        if 'ipantuserattrs' in self.get_attr_vals('objectClass'):
            # We need an update to ntUserPassword for this
            attribute_map.append( ('ipaNTHash', 'ntUserNtPassword') )
            attribute_map.append( ('krbprincipalname', 'ntUserDomainId') )
        # These map 1:1
        for (ipaattr, dsattr) in attribute_map:
            val = self.get_attr_vals(ipaattr)
            if val is not None and len(val) > 0:
                if dsattr == 'ntUserDomainId':
                    # we need to split this value is all
                    properties[dsattr] = val[0].split('@')[1]
                else:
                    properties[dsattr] = val

        # Returns a properties that can be passed to an UserAccounts
        # Map the ipa attr to the new properties
        return properties

class IpaUsers(DSLdapObjects):
    def __init__(self, instance, basedn, batch=False):
        super(IpaUsers, self).__init__(instance, batch)
        self._objectclasses = [
            'ipaobject',
            'person',
            'inetorgperson',
            'organizationalperson',
            'inetuser',
            'posixaccount',
            'ipaSshGroupOfPubKeys',
        ]
        self._filterattrs = ['uid']
        self._childobject = IpaUser
        self._basedn = 'cn=users,cn=accounts,%s' % basedn

class IpaGroup(DSLdapObject):
    def __init__(self, instance, dn=None, batch=False):
        super(IpaGroup, self).__init__(instance, dn, batch)
        self._rdn_attribute='cn'
        self._must_attributes = []
        self._create_objectclasses = [
            'ipaobject',
            'ipausergroup',
            'groupofnames',
            'posixgroup',
        ]
        self._protected = True

    def convert_to_posixgroup(self):
        # Returns a set of properties that can be given to the PosixGroups.create type
        properties = {}
        attribute_map = [
            ('cn', 'cn'),
            ('gidNumber', 'gidNumber'),
            ('description', 'description'),
        ]
        for (ipaattr, dsattr) in attribute_map:
            val = self.get_attr_vals(ipaattr)
            if val is not None and len(val) > 0:
                properties[dsattr] = val
        return properties

class IpaGroups(DSLdapObjects):
    def __init__(self, instance, basedn=None, batch=False):
        super(IpaGroups, self).__init__(instance, batch)
        self._objectclasses = [
            'ipaobject',
            'ipausergroup',
            'groupofnames',
            'posixgroup',
        ]
        self._filterattrs=['cn']
        self._childobject = IpaGroup
        self._basedn = 'cn=groups,cn=accounts,%s' % basedn


def _ipa_migrate_group(ipa_inst, inst, ipa_basedn, basedn, log, commit=False):
    # Which groups do we want to migrate?
    # Add an ipaGroup type
    # Give it an ipaGroup -> Group mapping ability
    ipagroups = IpaGroups(ipa_inst, ipa_basedn)
    dsgroups = PosixGroups(inst, basedn)
    for ipagroup in ipagroups.list():
        props = ipagroup.convert_to_posixgroup()
        log.debug('IPA group %s -> DS group properties %s' % (ipagroup.__unicode__(), props))
        try:
            rdn_value = ipagroup.get_attr_vals(ipagroup._rdn_attribute)[0]
            dsgroup = dsgroups.get(rdn_value)
            if commit is True:
                log.info("Syncing group data from %s to %s" % (ipagroup.__unicode__(), dsgroup.__unicode__()) )
                # ipauser.sync_to_useraccount(dsgroup)
            else:
                log.info("Would sync group data from %s to %s" % (ipagroup.__unicode__(), dsgroup.__unicode__()) )
        except ldap.NO_SUCH_OBJECT:
            if commit:
                log.info('Creating %s ...' % rdn_value)
                dsgroups.create(properties=props)
            else:
                log.info("Would create %s" % rdn_value)

    return True

def _ipa_migrate_users(ipa_inst, inst, ipa_basedn, basedn, log, commit=False):
    ipausers = IpaUsers(ipa_inst, ipa_basedn)
    dsusers = UserAccounts(inst, basedn)
    dsgroups = PosixGroups(inst, basedn)
    for ipauser in ipausers.list():
        log.debug(ipauser.__unicode__())
        props = ipauser.convert_to_posixaccount()
        log.debug('IPA user -> DS user properties %s' % props)
        try:
            # What happens if this is a multi valued attr?
            rdn_value = ipauser.get_attr_vals(ipauser._rdn_attribute)[0]
            dsuser = dsusers.get(rdn_value)
            # If the value doesn't exist the exception is thrown here.
            # if it does, we should sync the data
            if commit is True:
                log.info("Syncing user data from %s to %s" % (ipauser.__unicode__(), dsuser.__unicode__()) )
                # ipauser.sync_to_useraccount(dsuser)
            else:
                log.info("Would sync user data from %s to %s" % (ipauser.__unicode__(), dsuser.__unicode__()) )
        except ldap.NO_SUCH_OBJECT:
            if commit is True:
                log.info('Creating uid=%s,ou=People,%s' % (rdn_value, basedn))
                dsusers.create(properties=props)
            else:
                log.info('Would create uid=%s,ou=People,%s' % (rdn_value, basedn))
        # Now, check the group memberships
        # The user *should* exist at this point, unless we are just "checking"
        for member in ipauser.get_attr_vals('memberof'):
            # Explode the attr to a single rdn component
            name = member.split('=', 1)[1].split(',', 1)[0]

            log.info(member)
            if commit is True:
                dsuser = dsusers.get(rdn_value)
                log.debug('Checking %s in %s' % (dsuser.dn, name))

                try:
                    group = dsgroups.get(name)
                    ismember = group.check_member(dsuser.dn)
                    if ismember:
                        log.info('%s is already a member of %s' % (dsuser.dn, name))
                    elif not ismember:
                        log.info('Adding %s to %s' % (dsuser.dn, name))
                        group.add('member', dsuser.dn)

                except ldap.NO_SUCH_OBJECT:
                    log.warning('WARNING, unable to find group %s, skipping ...' % name)
            else:
                log.info('Would add %s to %s' % (ipauser.__unicode__(), name))
            # Do we do this by checking the posixGroup?
            # Or the memberOf?
            # I don't think we can guarantee memberOf is present ...


    return True


def _ipa_migrate(ipa_inst, inst, ipa_basedn, basedn, log, commit=False):

    if not _init(inst, basedn, log, commit=commit):
        log.error("Config migration check failed")
        return False

    if not _ipa_migrate_group(ipa_inst, inst, ipa_basedn, basedn, log, commit=commit):
        log.error("Group migration check failed")
        return False

    if not _ipa_migrate_users(ipa_inst, inst, ipa_basedn, basedn, log, commit=commit):
        log.error("Users migration check failed")
        return False

    return True

# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
