# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016, William Brown <william at blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import ldap

from lib389._mapped_object import DSLogging, _gen_and, _gen_filter, _term_gen
from lib389._entry import EntryAci

# These don't inherit from DSLdapObject* because aci is a bit ... different
# Because these are compound attributes, rather than true entries, we must
# handle them *very* differently.
# It's likely this will go back to lib389 one day.

class Aci(DSLogging):
    def __init__(self, instance, dn, rawaci=None, batch=False):
        """
        """
        self._instance = instance
        super(Aci, self).__init__(self._instance.verbose)
        # super(DSLdapObject, self).__init__(self._instance.verbose)
        # This allows some factor objects to be overriden
        if dn is not None:
            self._dn = dn
        else:
            raise ValueError('dn may NOT be None')

        self._batch = batch
        self._protected = False
        entry = self._instance.getEntry(self._dn)
        if rawaci is None:
            self._entryaci = EntryAci(entry, '', verbose=False)
        else:
            self._entryaci = EntryAci(entry, rawaci, verbose=False)
        self._rawaci = rawaci
        self._permission_create = False

    def __unicode__(self):
        # THIS should return the rawaci + dn.
        val = self._dn
        if self._rdn_attribute:
            # What if the rdn is multi value and we don't get the primary .... ARGHHH
            val = self.get_attr_val(self._rdn_attribute)
        return ensure_str(val)

    def __str__(self):
        return self.__unicode__()

    def _to_value(self, value):
        if isinstance(value, list):
            return [{'values': value}]
        elif isinstance(value, set):
            return [{'values': list(value)}]
        else:
            return [{'values': [value]}]

    def commit(self):
        # Get the entry
        # entry = self._instance.getEntry(self._dn)
        # Now format the rawaci
        new_rawaci = self._entryaci.getRawAci()
        # if _rawaci is None, just add
        self._log.debug('Commitng %s' % new_rawaci)
        if self._rawaci is not None:
            self._instance.modify_s(self._dn, [(ldap.MOD_DELETE, 'aci', self._rawaci)])
        self._instance.modify_s(self._dn, [(ldap.MOD_ADD, 'aci', new_rawaci)])
        self._rawaci = new_rawaci

    def set_aciname(self, name):
        self._entryaci.acidata['acl'] = self._to_value(name)

    def set_bindrule(self, bindrule):
        # Add the ldap:///self to the bindrules.
        self._entryaci.acidata['allow_raw_bindrules'] = self._to_value(bindrule)

    def create_permission(self, name):
        # Check and create the permission group
        # Set the raw bindrule to use it.
        pass

    def set_targetattr(self, attributes):
        self._entryaci.acidata['targetfilter'] = []
        self._entryaci.acidata['targetattr'] = self._to_value(set(attributes))

    # Default set of safe exclusions
    def set_targetobjectclass(self, objectclasses, attribute_exclude=['userPassword', 'ntuserntpassword']):
        # Will take the objectclasses from schema.
        rawattr = []
        for roc in objectclasses:
            oc = self._instance.schema.query_objectclass(roc)
            rawattr += oc.must
            rawattr += oc.may
        uniqattr = set(map(str.lower, rawattr))
        # We then REMOVE what's in the attribute exclude
        for attre in attribute_exclude:
            try:
                uniqattr.remove(attre.lower())
            except KeyError:
                # If it's not there, who cares!
                pass
        # It then generates the filter AND attribute lists.
        self._entryaci.acidata['targetattr'] = self._to_value(uniqattr)
        self._entryaci.acidata['targetfilter'] = self._to_value(_gen_and(
            _gen_filter(_term_gen('objectclass'), objectclasses)
            )
        )

        print(self._entryaci.acidata)

    def set_allow(self, value):
        self._entryaci.acidata['allow'] = self._to_value(value)

    def match(self, key, value):
        # Does this aci match some selector component.
        data = []
        if key == 'allow':
            data = self._entryaci.acidata['allow']
        elif key == 'targetattr':
            data = self._entryaci.acidata['targetattr']
        elif key == 'desc':
            data = self._entryaci.acidata['acl']
        elif key == 'bindrule':
            data = self._entryaci.acidata['allow_raw_bindrules']
        for datum in data:
            for v in datum['values']:
                if value in v:
                    return True
        return False

    def delete(self):
        if self._protected:
            raise ldap.UNWILLING_TO_PERFORM('Cannot delete protected object')

        self._instance.modify_s(self._dn, [(ldap.MOD_DELETE, 'aci', self._rawaci)])

class Acis(DSLogging):
    def __init__(self, instance, basedn, batch=False):
        self._childobject = Aci
        self._instance = instance
        super(Acis, self).__init__(self._instance.verbose)
        # super(DSLdapObjects, self).__init__(self._instance.verbose)
        # self._objectclasses = []
        # self._filterattrs = []
        # self._list_attrlist = ['dn']
        # Copy this from the child if we need.
        self._basedn = basedn
        self._batch = batch
        self._scope = ldap.SCOPE_SUBTREE

    def list(self):
        acis = []
        rawacientries = self._instance.search_s(self._basedn, self._scope, 'aci=*', ['aci'])
        for rawacientry in rawacientries:
            # This will become the Aci type.
            for aci in rawacientry.getValues('aci'):
                acis.append( Aci(self._instance, rawacientry.dn, aci) )
        return acis


    def search(self, key, value):
        acis = self.list()
        candidates = []
        for aci in acis:
            if aci.match(key, value):
                candidates.append(aci)
        # Get the list of acis
        # Now map over them with the selector
        # We can't do this in an ldap search because we rely on the aci components
        return candidates

    def purge(self):
        for aci in self.list():
            self._log.debug('purging %s %s' % (aci._dn, aci._rawaci))
            aci.delete()


def create_base_acis():
    # Create the basic aci's needed for idm389
    # Self read? All read to some extent? Bind?
    pass


# Are permissions actually the aci string? Or are they the group?

def create_permission():
    # Create and ACI and the related permission group
    pass


def remove_permission():
    # Remove the permission and it's related group
    pass

def add_user_permission():
    pass

def remove_user_permission():
    pass
